.connectionstatus
{
    z-index: 1000;
    width: 100%;
    position: absolute;
    top: 20px;
    .alert {
        margin-left: auto;
        margin-right: auto;
        width: 300px;
        text-align: center;
    }
}
