.bb-builders-view-form-control {
  display: block;
  width: 100%;
  height: 34px;
  padding: 6px 12px;
  margin-bottom: 20px;
  font-size: 14px;
  line-height: 1.42857143;
  color: #555555;
  background-color: #fff;
  border: 1px solid #ccc;
  border-radius: 4px;
  box-shadow: inset 0 1px 1px rgba(0, 0, 0, 0.075);
  transition: border-color ease-in-out .15s, box-shadow ease-in-out .15s;
}

.bb-builders-view-container {
  padding-right: 15px;
  padding-left: 15px;
  margin-left: 0;
  margin-right: 0;
}

.bb-builder-tag {
  padding: 0.5em 0.6em 0.3em 0.6em;
  font-size: 75%;
  font-weight: 700;
  line-height: 1;
  color: #fff;
  background-color: #777777;
  text-align: center;
  white-space: nowrap;
  border-radius: 0.25em;
}

.bb-builder-tag-filtered {
  background-color: #5cb85c;
}

.bb-builders-view-help-popover {
  &.popover {
    padding: 1px;
    border-radius: 6px;
    box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
    margin-top: 10px;
  }

  .popover-title {
    padding: 8px 14px;
    margin: 0;
    font-size: 14px;
    background-color: #f7f7f7;
    border-bottom: 1px solid #ebebeb;
    border-radius: 5px 5px 0 0;
  }

  .popover-content {
    padding: 9px 14px;
  }

  pre {
    display: block;
    padding: 9.5px;
    margin: 0 0 10px;
    font-size: 13px;
    word-break: break-all;
    word-wrap: break-word;
    background-color: #f5f5f5;
    border: 1px solid #ccc;
    border-radius: 4px;
  }
}
