html {
  overflow-y: scroll;
}
html, body, #root {
  height: 100%;
}
.row {
  margin-left: 0;
  margin-right: 0;
}
.row > div {
  margin-bottom: 15px;
}


.noselect {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -khtml-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}
