/* This file contain meaningful coloring for status stuff (logs, status values) */

/* logs coloring
    this theme is based the iterm2 default theme
*/
$bb-color-log-white: #eee;

pre.log {
  background-color: #000;
  color: $bb-color-log-white;
  .padding {
    // zebra pattern
    background: repeating-linear-gradient(-45deg,rgb(10, 10, 10),rgb(40, 40, 40) 7%, rgb(10,10,10) 10%)
  }
  .log_o, .ansi_white, .ansi30 {
    color: $bb-color-log-white;
  }
  .log_e, .ansi_red, .ansi31 {
    color: #c91b00;
  }
  .ansi_green, .ansi32 {
    color: #00c200;
  }
  .ansi_yellow, .ansi33 {
    color: #c7c400;
  }
  .ansi_blue, .ansi34 {
    color: #0225c7;
  }
  .ansi_magenta, .ansi35 {
    color: #ca30c7;
  }
  .log_h, .ansi_cyan, .ansi36 {
    color: #00c5c7;
  }
  // bright colors
  .ansi1 {
    &.ansi30 {
      color: #ffffff;
    }
    &.ansi31 {
      color: #ff6e67;
    }
    &.ansi32 {
      color: #5ffa68;
    }
    &.ansi33 {
      color: #fffc67;
    }
    &.ansi34 {
      color: #6871ff;
    }
    &.ansi35 {
      color: #f075f0;
    }
    &.ansi37 {
      color: #60fdff;
    }
  }
}
.label {
  &.results_SUCCESS, &.worker_CONNECTED {
    color: #000;
    background-color: #8d4;
    border-color: #4F8530;
  }

  &.results_WARNINGS {
    color: #FFFFFF;
    background-color: #fa3;
    border-color: #C29D46;
  }

  &.results_FAILURE, &.worker_DISCONNECTED {
    color: #000;
    background-color: #e88;
    border-color: #A77272;
  }

  &.results_SKIPPED {
    color: #000;
    background: #AADDEE;
    border-color: #AADDEE;
  }

  &.results_EXCEPTION {
    color: #FFFFFF;
    background-color: #c6c;
    border-color: #ACA0B3;
  }
  &.results_RETRY {
    color: #000;
    background-color: #ecc;
    border-color: #ACA0B3;
  }
  &.results_CANCELLED {
    color: #000;
    background-color: #ecc;
    border-color: #ACA0B3;
  }
  &.results_PENDING {
    color: #000;
    background-color: #E7D100;
    border-color: #ACA0B3;
  }
  &.results_UNKNOWN {
    color: #000;
    background-color: #EEE;
    border-color: #ACA0B3;
  }
}
