
@import "../module/**/*.less";

@column-width: 40px;

.console {
    .table-fixedwidth {
        width: initial;
    }
    .load-indicator {
        width: 100%;
        height: 100%;
        z-index: 900;
        background-color: #ffffff;
        display: table;

        .spinner {
            display: table-cell;
            vertical-align: middle;
            text-align: center;

            p {
                font-weight: 300;
                margin-top: 10px;
            }
        }
    }

    .column {
        min-width: @column-width;
        max-width: @column-width;
        width: @column-width;
    }
    table {
        border: none;
    }
    .tag_row{
        td {
            margin:0px;
            padding:0px;
        }
        span {
            position: relative;
            float: left;
            font-size: 10px;
            overflow: hidden;
            text-decoration: none;
            white-space: nowrap;
        }
    }
    tr.first-row {
        background-color: #fff!important;
        th {
        border: none;
        background-color: #fff !important;
        }
        .builder {
            position: relative;
            float: left;
            font-size: 12px;
            text-align: center;
            transform: rotate(-25deg) ;
            transform-origin: 0% 100%;
            text-decoration: none;
            white-space: nowrap;
        }
    }

}
