
.bb-properties-value {
  vertical-align: top;
  padding: unset;
  margin: unset;
  display: inline-block;
  border: unset;
  background-color: unset;
}

.bb-properties-copy {
  vertical-align: top;
  padding: 0 0.5ex;
}
