/* mobile tweaks

  Reduce a bit more the font sizes and paddings the make more stuff in the screen, trading a bit of uglyness...
*/
@media (max-width: 600px) {
  .navbar .breadcrumb {
    font-size: 12px !important;;
  }
  .container, .container-fluid, .panel-heading, .list-group-item {
    padding-left: 5px !important;
    padding-right: 5px !important;
  }
  .navbar {
    .container, .container-fluid {
      padding-left: 15px !important;
      padding-right: 15px !important;
    }
  }
  .nav.nav-tabs > li > a {
    padding: 5px !important;
    font-size: 12px !important;
  }
  .logline {
    height: 12px !important;;
    font-size: 10px !important;
  }
  .nomobile{ display: None}
}
