
/* The RTD default uses a lot of horizontal space for the "Parameters" heading which is placed
    to a separate column. The descriptions of parameters are often very long in Buildbot, so
    just some indentation is fine
*/

html.writer-html5 .rst-content dl.field-list {
    display: initial;
}

html.writer-html5 .rst-content dl.field-list dt {
    padding-left: 0;
}

html.writer-html5 .rst-content dl.field-list dd {
    margin-left: 0;
}

/* The left and right padding for the contents seems excessive
*/

.wy-nav-content {
    /* the values used are the same as for top and bottom padding */
    padding-right: 1.618em;
    padding-left: 1.618em;
}

/* The vertical margin seems excessive on descriptions of parameters when there are many
   parameter sets. The structure of the page becomes visually unclear.
*/
.rst-content dl dd,
.rst-content dl dt,
.rst-content dl ol,
.rst-content dl p,
.rst-content dl table,
.rst-content dl ul {
    margin-bottom: 4px;
}

/* The default is one line per paragraph spacing which seems excessive. */
p {
    margin-bottom: 8px;
}
