.fade.modal.show {
  // Otherwise the modal is not shown.
  opacity: 1;
}

@media screen {
  .modal-dialog {
    max-width: 100%;
    width: 90%;
  }

  .modal-body {
    overflow-y: auto;
    max-height: 80vh;
  }
}
