.bb-buildsticker {
  width: 100%;
  margin: 0 2px 2px 0;

  .bb-buildsticker-left {
    white-space: nowrap;
    text-overflow: ellipsis;
    overflow: hidden;
    margin-left: 0;
    margin-right: 0;
  }

  &.results_SUCCESS {
    border-color: #4F8530;
  }

  &.results_WARNINGS {
    border-color: #C29D46;
  }

  &.results_FAILURE {
    border-color: #A77272;
  }

  &.results_SKIPPED {
    border-color: #AADDEE;
  }

  &.results_EXCEPTION {
    border-color: #ACA0B3;
  }

  &.results_RETRY {
    border-color: #ACA0B3;
  }

  &.results_CANCELLED {
    border-color: #ACA0B3;
  }

  &.results_PENDING {
    border-color: #ACA0B3;
  }

  &.results_UNKNOWN {
    border-color: #ACA0B3;
  }
}
