.bb-badge-status {
  display: inline;
  padding: 0.2em 0.6em 0.3em;
  font-size: 75%;
  font-weight: 700;
  line-height: 1;
  text-align: center;
  white-space: nowrap;
  vertical-align: baseline;
  border-radius: 0.25em;

  &.results_SUCCESS, &.worker_CONNECTED {
    color: #000;
    background-color: #8d4;
    border-color: #4F8530;
  }

  &.results_WARNINGS {
    color: #FFFFFF;
    background-color: #fa3;
    border-color: #C29D46;
  }

  &.results_FAILURE, &.worker_DISCONNECTED {
    color: #000;
    background-color: #e88;
    border-color: #A77272;
  }

  &.results_SKIPPED {
    color: #000;
    background: #AADDEE;
    border-color: #AADDEE;
  }

  &.results_EXCEPTION {
    color: #FFFFFF;
    background-color: #c6c;
    border-color: #ACA0B3;
  }

  &.results_RETRY {
    color: #000;
    background-color: #ecc;
    border-color: #ACA0B3;
  }

  &.results_CANCELLED {
    color: #000;
    background-color: #ecc;
    border-color: #ACA0B3;
  }

  &.results_PENDING {
    color: #000;
    background-color: #E7D100;
    border-color: #ACA0B3;
  }

  &.results_UNKNOWN {
    color: #000;
    background-color: #EEE;
    border-color: #ACA0B3;
  }
}

a .bb-badge-status:hover,
a .bb-badge-status:focus {
  color: #fff;
  cursor: pointer;
}
