
.change-avatar {
  background-size: 16px 16px;
  border-radius: 50%;
  display: inline-block;
  margin-bottom: -3px;
  margin-right: 10px;
  height: 20px;
  width: 20px;
  overflow: hidden;
  img {
    height: 20px;
    width: 20px;
    background-color: #ccc
  }
}
